# Instruction benchmarking

We automatically generate Hexagon assembly for every instruction (or as many as we can), that simply calls that instruction multiple times. This allows for benchmarking on the granularity of single instructions, and can hopefully aid in rooting out performance problems in `qemu-hexagon`.

As an example, for `A2_add`, we emit
```
#define SYS_exit_group 94
    .text
    .globl _start
_start:
    allocframe(#32)
    // Loop prologue
    r0 = #0
    r1 = #1000000
    loop:
    // Loop iteration
    r3 = #0x1
    r4 = #0x1
    r2=add(r3,r4)
    r2=add(r3,r4)
    r2=add(r3,r4)
    ...
    r2=add(r3,r4)
    r2=add(r3,r4)
    r2=add(r3,r4)
    // Loop epilogue
    r0 = add(r0, #1)
    p0 = cmp.lt(r0, r1)
    if (p0) jump loop
    // Exit
    deallocframe
    r0 = #0
    r6 = #SYS_exit_group
    trap0(#1)
```

# Overview of files

Overview of the various files in this project, note building, running, and plotting are covered more in depth later on.
* `build.sh` Runs `gen_semantics.c` and `emit_bench_asm.py` to emit assembly code into `build/asm`, compiles the assembly code into Hexagon executables in `build/bin`
* `run.sh` Runs and times all the code in `build/bin` for a list of QEMU executables, dumps the output to stdout
* `src/emit_bench_asm.py` The script actually responsible for emitting the assembly code
* `src/plot.jl` Plots the benchmark output from `run.sh` using `Julia`
* `src/hex_common.py` Taken from `qemu-hexagon`
* `src/gen_semantics.c` Taken from `qemu-hexagon`
* `src/imported`  Taken from `qemu-hexagon`

# Building

The `build.sh` script takes a few arguments
```shell
> ./build.sh /path/to/python /path/to/clang /path/to/hexagon-clang
```
where `clang` is used to compile `src/gen_semantics.c`, `python` used to invoke `src/emit_bench_asm.py`, and `hexagon-clang` used to compile the generated assembly code into Hexagon executables.

We usually rely on docker container provided by QEMU for hexagon-clang, and so
```shell
> QEMU_DIR="/path/to/qemu"
> HEXAGON_CLANG="${QEMU_DIR}/tests/docker/docker.py --engine auto cc --cc hexagon-unknown-linux-musl-clang -i qemu/debian-hexagon-cross -s ${QEMU_DIR} --"
> ./build.sh python clang $HEXAGON_CLANG
```
works perfectly fine. Running through docker might take a while (~16 mins on a 4 core i5) since there are 1036 instructions for which we currently generate assembly and then compile.

After `build.sh` is finished you should have a `build` directory containing `build/asm` and `build/bin` with the assembly and executable code respectively.

There are few tunable parameters in `build.sh`, namely
* `num_loop_iterations` Amount of loop iterations to perform (default 1000000), looping is disabled if this is set to `0`,
* `insn_per_iteration` Amount of instructions to emit per loop iteration (default 100).
The defaults are chosen so the elapsed time per instructions is at least in the order of milliseconds for the fastest instructions.

# Running

At this point you can of course run the individual tests manually in `build/bin`, but `run.sh` does simplify this considerably, especially when multiple versions of QEMU are involved. The script takes a variable amount of arguments but you have to provide a label and path for every version of QEMU you want to test, for instance (you can of course exclude the `column` pipe, it's just there for the sake of prettier output)
```shell
> ./run.sh "phase1" /path/to/phase1 \
           "phase2" /path/to/phase2 \
           "qemu"   /path/to/upstream \
           "quic"   /path/to/quic \
           | column -t -s ' ' \
           > bench_out
```
will give us the following in `bench_out`
```
> cat bench_out
Instruction         phase1  phase2  qemu   quic
A2_abs              0.034   0.025   0.165  0.010
A2_absp             0.027   0.027   0.232  0.197
A2_abssat           0.056   0.026   0.225  0.197
A2_add              0.026   0.026   0.168  0.010
A2_addh_h16_hh      0.026   0.026   0.182  0.142
A2_addh_h16_hl      0.027   0.026   0.171  0.135
A2_addh_h16_lh      0.026   0.026   0.170  0.140
A2_addh_h16_ll      0.027   0.027   0.169  0.138
A2_addh_h16_sat_hh  0.056   0.027   0.214  0.206
A2_addh_h16_sat_hl  0.056   0.026   0.235  0.198
A2_addh_h16_sat_lh  0.057   0.027   0.229  0.195
...
```

Note `run.sh` will also take some time to run for all tests (~18 mins). As it stands currently, 46 tests will crash and not be included in the final output of `run.sh`, the reason for these crashes is that they use immediate values for pointers and `emit_bench_asm.py` does not yet emit valid immediate-valued pointers.

# Plotting

Requires `julia` with the `Plots, DelimitedFiles, Distributions, Printf` packages and is run as follows
```shell
> julia src/plot.jl bench_out "qemu"
```
the first argument specifies where to find the benchmark data, and the second specifies the label of the column to use for reference data. The reference data will be normalized against in two of the plots. A total of three plots should be produced:
1. Plot of raw data for each version of QEMU
2. Plot of normalized data (against reference data specified by label) for each version of QEMU
3. Plot of normalized data along with fitted normal distrubutions to get an idea of how the timing results distribute

We currently use the `plotly` backend which will open interactive graphs in your closest web browser!

# Limitations

We don't handle:
* instruction with immediate pointer arguments,
* HVX instructions.
