import sys
import re
import os

import hex_common

# Helper class to make it a bit easier to refer to
# the different parts of a register
class HexagonReg:
    def __init__(self, regtype, width, name, decoration):
        self.type = regtype
        self.width = width
        self.name = name
        self.decoration = decoration

# class that makes it a bit easier to emit Hexagon asm,
# keeps track of which registers have been allocated for
# every register type
class HexagonAsmWriter:
    def __init__(self, output_file):
        self.output_file = output_file
        self.indent_level = 0
        self.index_map = {"r" : 0, "p" : 0, "m" : 0, "v" : 0, "q" : 0}

    def set_indent_level(self, indent_level):
        self.indent_level = indent_level

    def write(self, string):
        self.output_file.write(string)

    def write_line(self, line):
        self.write(" "*self.indent_level + line + "\n")

    def new_register(self, regtype, width, toss="", conj=""):
        assert(width == 32 or width == 64 or
               width == 128 or width == 256)
        regtype = regtype.lower()
        indices = []

        if width == 32 or width == 128:
            indices.append(self.index_map[regtype])
            self.index_map[regtype] += 1
        else:
            # For register pairs Hexagon requires the lower-half register
            # to have an even index, so here we skip over a register if
            # it's not even
            if self.index_map[regtype] % 2 != 0:
                self.index_map[regtype] += 1
            indices.append(self.index_map[regtype])
            self.index_map[regtype] += 1
            indices.append(self.index_map[regtype])
            self.index_map[regtype] += 1

        name = "%s%s" % (regtype, ":".join(map(str, reversed(indices))))
        decoration = toss + conj

        return HexagonReg(regtype, width, name, decoration)

    def clear_registers(self):
        self.index_map = {"r" : 0, "p" : 0, "m" : 0, "v" : 0, "q" : 0}

    def begin_packet(self):
        self.write_line("{")
        self.indent_level += 4

    def end_packet(self):
        self.indent_level -= 4
        self.write_line("}")

allowed_insn_beh = re.compile(r"([+|-]?=)(.*)\(.*\)(.*)")
def should_skip_insn(tag):
    # Skip jump, and floating point instructions
    if tag.startswith("J") or tag.startswith("F2"):
        return True

    # Skip .new instructions
    if ".new" in hex_common.behdict[tag]:
        return True

    if not allowed_insn_beh.findall(hex_common.behdict[tag]):
        return True

    return False

def get_reg_width(regtype, regid):
    if (regtype == "V" or regtype == "Q"):
        return 128 if hex_common.is_single(regid) else 256
    else:
        return 32 if hex_common.is_single(regid) else 64

def main():
    if len(sys.argv) < 3:
        print("usage: python emit_bench_asm.py [semantics file] [output directory] [num loop iterations] [insn per iteration]")
        return

    semantics          = sys.argv[1]
    output_dir         = sys.argv[2]
    loop_iterations    = int(sys.argv[3])
    insn_per_iteration = int(sys.argv[4])

    hex_common.read_semantics_file(semantics)
    tagregs = hex_common.get_tagregs()
    tagimms = hex_common.get_tagimms()

    for tag in hex_common.tags:
        if should_skip_insn(tag):
            continue

        with open("%s/%s.S" % (output_dir, tag), "w") as output_file:
            asm = HexagonAsmWriter(output_file)

            asm.set_indent_level(0)
            asm.write_line("#define SYS_exit_group 94")
            asm.set_indent_level(4)
            asm.write_line(".text")
            asm.write_line(".globl _start")
            asm.set_indent_level(0)
            asm.write_line("_start:")
            asm.set_indent_level(4)

            asm.write_line("allocframe(#32)")

            loop_count = None
            loop_count_max = None
            # Emit loop prologue if needed
            if loop_iterations > 1:
                loop_count = asm.new_register("r", 32)
                loop_count_max = asm.new_register("r", 32)
                asm.write_line("// Loop prologue")
                asm.write_line("%s = #0" % loop_count.name)
                asm.write_line("%s = #%s" % (loop_count_max.name, loop_iterations))
                asm.write_line("loop:")
                asm.write_line("// Loop iteration")

            insn_string = hex_common.behdict[tag]

            # Loop over and allocate a gpr register for each register the
            # instruction uses
            #
            # read_regs keeps track of registers that are read from, these
            # registers will have to be initialized to some reasonable
            # value
            read_regs = []
            for regtype, regid, toss, numregs, conj in tagregs[tag]:
                width = get_reg_width(regtype, regid)
                reg = asm.new_register(regtype, width, toss, conj)

                # Replace register string in insn_string with the register we allocated for it
                insn_string = insn_string.replace(regtype + regid + toss + numregs + conj, reg.name + reg.decoration)

                assert(hex_common.is_read(regid) or hex_common.is_written(regid))

                if hex_common.is_read(regid):
                    read_regs.append(reg)

            # Replace immediates in insn_string with values
            for immlett, bits, immshift in tagimms[tag]:
                immstring = "#" + immlett + bits
                if immshift:
                    immstring += ":" + immshift
                    insn_string = insn_string.replace(immstring, "#" + str(1 << int(immshift)))
                else:
                    insn_string = insn_string.replace(immstring, "#0x1")

            # We need to handle instructions that expect pointers in its
            # register arguments differently, which is why we check for
            # load/store instructions. Note A_RESTRICT_SLOT0ONLY is to
            # also handle mem-ops instructions.
            isloadstore = "A_LOAD"  in hex_common.attribdict[tag] or \
                          "A_STORE" in hex_common.attribdict[tag] or \
                          "A_RESTRICT_SLOT0ONLY" in hex_common.attribdict[tag]

            # Here we loop over all registers the instruction reads from
            # and try to give them sensible initial values
            for reg in read_regs:
                if reg.type == "r":
                    value = "sp" if isloadstore else "#0x1"
                    if reg.width == 64:
                        asm.write_line("%s = combine(#0x0, %s)" % (reg.name, value))
                    else:
                        asm.write_line("%s = %s" % (reg.name, value))
                elif reg.type == "p":
                    asm.write_line("%s = cmp.eq(r0,r0)" % (reg.name))
                elif reg.type == "m":
                    tmp = asm.new_register("r", 32)
                    asm.write_line("%s = #0x0" % tmp.name)
                    asm.write_line("%s = %s" % (reg.name, tmp.name))
                elif reg.type == "v":
                    if reg.width == 128:
                        value = asm.new_register("r", 32)
                        asm.write_line("%s = #0x1" % value.name)
                        asm.write_line("%s = vsplat(%s)" % (reg.name, value.name))
                    else:
                        value = asm.new_register("r", 32)
                        splat = asm.new_register("v", 128)
                        asm.write_line("%s = #0x1" % value.name)
                        asm.write_line("%s = vsplat(%s)" % (splat.name, value.name))
                        asm.write_line("%s = vcombine(%s, %s)" % (reg.name, splat.name, splat.name))
                elif reg.type == "q":
                    value = asm.new_register("r", 32)
                    splat = asm.new_register("v", 128)
                    assert(reg.width == 128)
                    asm.write_line("%s = #0x0" % value.name)
                    asm.write_line("%s = vsplat(%s)" % (splat.name, value.name))
                    asm.write_line("%s = vcmp.eq(%s.w, %s.w)" % (reg.name, splat.name, splat.name))
                else:
                    print("Unknown reg type: %s for %s\n" % (reg.type, tag))
                    assert(False)

            # For instructions that uses gp we just initialize it to the stack pointer
            if isloadstore and "gp" in hex_common.behdict[tag]:
                asm.write_line("gp = sp")

            # Emit actual instruction code
            for i in range(0, insn_per_iteration):
                asm.write_line("%s" % (insn_string))

            # Emit loop epilogue if needed
            if loop_iterations > 1:
                asm.write_line("// Loop epilogue")
                asm.write_line("%s = add(%s, #1)" % (loop_count.name, loop_count.name))
                asm.write_line("p0 = cmp.lt(%s, %s)" % (loop_count.name, loop_count_max.name))
                asm.write_line("if (p0) jump loop")

            asm.write_line("// Exit")
            asm.write_line("deallocframe")
            asm.write_line("r0 = #0")
            asm.write_line("r6 = #SYS_exit_group")
            asm.write_line("trap0(#1)")

if __name__ == "__main__":
    main()
