using Plots, DelimitedFiles, Distributions, Printf
plotly()

if length(ARGS) != 2
    println("usage: julia plot.jl [benchmark output file] [label in benchmark file of reference data]")
    exit
end

bench_file = ARGS[1]
reference_label = ARGS[2]

raw_data = readdlm(bench_file)
header   = raw_data[1, 2:end]
insn     = raw_data[2:end, 1]
data     = raw_data[2:end, 2:end]

#
# 1st plot.
# Loop over and plot raw data for each column in the benchmark file
#

# For the x-coordinate we just use the row of the data
x = 1:size(data,1)
p1 = plot(size=(1920,1080), xlabel="Instruction", ylabel="Time (ms)", yticks=0:1e2:4e4)
for i = 1:size(data,2)
    # Note we plot in ms here, thereby the .*1000 as the input is seconds
    plot!(x, data[:,i].*1000, markershape=:circle, hover=insn, label=header[i])
end
display(p1)

# We now wish to normalize the timing data against some reference data,
# (e.g. upstream qemu).
reference_index = findfirst(x -> x == reference_label, header)
if isnothing(reference_index)
    println("Error: Reference label ", reference_label, " not found in benchmark file ", bench_file)
    exit
end

# Normalize data against reference data
normalized_data = data[:, 1:end .!= reference_index] ./ data[:, reference_index]
normalized_header = header[1:end .!= reference_index]

#
# 2nd plot.
# Loop over and plot normalized data
#

relative_time_label = (@sprintf "Time relative to %s, lower is better (%%)" reference_label)
p2 = plot(size=(1920,1080), xlabel="Instruction", ylabel=relative_time_label, yticks=0:0.1:3)
for i = 1:size(normalized_data,2)
    plot!(x, normalized_data[:,i], markershape=:circle, hover=insn, label=normalized_header[i])
end
display(p2)

#
# 3rd plot.
#
# We loop over the normalized data once more, this time however we fit a
# normal distribution to the data. The idea is to get a better sense of
# the distribution of timing results.
#

x = 0:0.1:3.5;
p3 = plot(layout=(3,1), size=(1080,1080), ylim=(-0.025,1.5), ylabel="Density")
# Set xlabel only for the 3rd subplot
plot!(subplot=3, xlabel=relative_time_label)
subplot_num = 1
for i = 1:size(normalized_data,2)
    y = normalized_data[:,i]
    n = fit(Normal, y)
    # Plot a vertical line at the x=1 mark
    vline!([1],       linecolor=:grey, linealpha=0.75, line=(:dash, 2),  label="", subplot=subplot_num)
    # Plot a vertical line at x=mu mean value
    vline!([mean(n)], linecolor=:grey, linealpha=0.75, line=(:solid, 2), label="", subplot=subplot_num, xticks=([0, mean(n), 1, 2], ["0", (@sprintf "μ= %.2f" mean(n)), "1", "2"]))
    # Plot a solid line for the actual normal distribution
    plot!(x, map(t -> pdf(n, t), x), linecolor=subplot_num, subplot=subplot_num, hover="", title=normalized_header[i], label="")
    # Scatter plot the data points on the normal distribution curve
    scatter!(y, map(t -> pdf(n, t), y), markeralpha=0.25, markercolor=subplot_num, subplot=subplot_num, label="", hover=insn)

    global subplot_num += 1
end
display(p3)
